#include <windows.h>
#include <windowsx.h>
#include "ddraw.h"
#include "d3drm.h"
#include "d3drmwin.h"

BOOL CreateScene();
void DestroyScene();
void ScaleMesh(LPDIRECT3DRMMESHBUILDER mesh, D3DVALUE dim);

int yes_full, yes_render, yes_hardware;
HWND hwnd;

LPDIRECT3DRM rm=0;
LPDIRECTDRAWCLIPPER clip=0;
LPDIRECT3DRMDEVICE device=0;
LPDIRECT3DRMFRAME scene=0;
LPDIRECTDRAW ddraw=0;
LPDIRECTDRAWSURFACE primsurf=0;
LPDIRECTDRAWSURFACE backsurf=0;
LPDIRECTDRAWSURFACE zbufsurf=0;
LPDIRECT3DRMFRAME camera;
LPDIRECT3DRMVIEWPORT viewport;
LPDIRECT3DRMMESHBUILDER meshbuilder;

D3DDEVICEDESC DrvDesc;

RECT rect_win;
int bpp;

GUID* ptr_guid=0;

#define MAX_DRV 5

struct {
	GUID	Guid;
	char	Name[50];
	char	About[50];
	BOOL	isHardware;
	D3DDEVICEDESC	Desc;
} DrvInfo[MAX_DRV];

int ptr_drv=0;


static DWORD BPPToDDBD(int bpp)
{
	switch(bpp) 
	{
	case 1:		return DDBD_1;
	case 2:		return DDBD_2;
	case 4:		return DDBD_4;
	case 8:		return DDBD_8;
	case 16:	return DDBD_16;
	case 24:	return DDBD_24;
	case 32:	return DDBD_32;
	default:	return 0;
	}
}


static HRESULT WINAPI enumDeviceFunc(
	LPGUID lpGuid, LPSTR lpDeviceDescription,
    LPSTR lpDeviceName, LPD3DDEVICEDESC lpHWDesc,
    LPD3DDEVICEDESC lpHELDesc, LPVOID lpContext )
{
	if ( ptr_drv>=MAX_DRV ) return D3DENUMRET_CANCEL;

    memcpy(&DrvInfo[ptr_drv].Guid, lpGuid, sizeof(GUID));
    lstrcpy(DrvInfo[ptr_drv].About, lpDeviceDescription);
    lstrcpy(DrvInfo[ptr_drv].Name, lpDeviceName);
	if ( lpHWDesc->dcmColorModel )
	{
		DrvInfo[ptr_drv].isHardware=1;
		memcpy(&DrvInfo[ptr_drv].Desc, lpHWDesc,sizeof(D3DDEVICEDESC));
	}
	else
	{
		DrvInfo[ptr_drv].isHardware=0;
		memcpy(&DrvInfo[ptr_drv].Desc, lpHELDesc, sizeof(D3DDEVICEDESC));
	}

	ptr_drv++;

	return D3DENUMRET_OK;
}


GUID* FindBestGUID(int ptr_bpp, int hw, LPD3DDEVICEDESC lpDesc)
{
	GUID* lpguid=0;
	LPDIRECTDRAW ddraw;
	LPDIRECT3D d3d;
	HRESULT rc;
	int i;

	rc = DirectDrawCreate( 0, &ddraw, 0 );
	if (rc!=DD_OK)
		return lpguid;

	rc = ddraw->QueryInterface( IID_IDirect3D, (void**)&d3d );
	if ( rc != D3DRM_OK )
		return lpguid;

	rc=d3d->EnumDevices( &enumDeviceFunc, 0 );
	if ( rc != D3D_OK )
		return lpguid;

	d3d->Release();
	ddraw->Release();

	for ( i=0; i<ptr_drv; i++ )
	{
		if ( DrvInfo[i].isHardware==hw )
		if ( DrvInfo[i].Desc.dwDeviceRenderBitDepth & BPPToDDBD(ptr_bpp))
		if ( DrvInfo[i].Desc.dcmColorModel==D3DCOLOR_RGB )
		{
			lpDesc=&DrvInfo[i].Desc;
			return &DrvInfo[i].Guid;
		}
	}
	return lpguid;
}



BOOL	InitD3DRM(HWND hWnd, int full_win, int wid, int hei, int bit, int yes_hw )
{
	hwnd=hWnd;
	yes_full=full_win;
	yes_hardware=yes_hw;

	GetClientRect(hwnd,&rect_win);

	HDC hdc=GetDC(hwnd);
	bpp=GetDeviceCaps(hdc,BITSPIXEL);
	ReleaseDC(hwnd,hdc);

	ptr_guid=FindBestGUID ( yes_full==1?bit:bpp, yes_hardware, &DrvDesc);
	if ( ptr_guid==0 ) yes_hardware=0;

	HRESULT rc;
	
	rc=Direct3DRMCreate (&rm);
	if ( rc!=D3DRM_OK ) return FALSE;

	if ( yes_full==0 )
	{
		rc=DirectDrawCreateClipper(0,&clip,NULL);
		if ( rc!=D3DRM_OK ) return FALSE;
		clip->SetHWnd(NULL,hwnd);
		if ( rc!=DD_OK ) return FALSE;

		rc=rm->CreateDeviceFromClipper( clip, ptr_guid, rect_win.right, rect_win.bottom, &device );
		if ( rc!=D3DRM_OK )
		{
			if ( clip ) { clip->Release(); clip=0; }
			if ( rm ) { rm->Release(); rm=0; }
			return FALSE;
		}
	}
	else
	{
		ShowCursor(FALSE);
		DirectDrawCreate(0,&ddraw,0);
		ddraw->SetCooperativeLevel( hwnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN );
		ddraw->SetDisplayMode( wid, hei, bit );
		rect_win.right=wid;
		rect_win.bottom=hei;

		DDSURFACEDESC desc;
		desc.dwSize = sizeof(desc);
		desc.dwFlags = DDSD_BACKBUFFERCOUNT | DDSD_CAPS;
		desc.dwBackBufferCount = 1;
		desc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_3DDEVICE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
		ddraw->CreateSurface(&desc,&primsurf,0);

		DDSCAPS ddscaps;
		ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
		primsurf->GetAttachedSurface(&ddscaps,&backsurf);

		memset(&desc,0,sizeof(desc));
		desc.dwSize = sizeof(DDSURFACEDESC);
		desc.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS | DDSD_ZBUFFERBITDEPTH;
		desc.dwWidth = wid;
		desc.dwHeight = hei;
		int bit_z;
		if ( yes_hardware )
		{
			int devDepth = DrvDesc.dwDeviceZBufferBitDepth;
			if (devDepth & DDBD_32)
				bit_z = 32;
			else if (devDepth & DDBD_24)
				bit_z = 24;
			else if (devDepth & DDBD_16)
				bit_z = 16;
			else if (devDepth & DDBD_8)
				bit_z = 8;
		}
		else	bit_z=16;
    	desc.dwZBufferBitDepth = bit_z;
		desc.ddsCaps.dwCaps = DDSCAPS_ZBUFFER;
		if ( yes_hardware )
			desc.ddsCaps.dwCaps |= DDSCAPS_VIDEOMEMORY;
		else
			desc.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;
		ddraw->CreateSurface(&desc,&zbufsurf,0);
		backsurf->AddAttachedSurface(zbufsurf);

		rc=rm->CreateDeviceFromSurface(ptr_guid,ddraw,backsurf,&device);
		if ( rc!=D3DRM_OK )
		{
			ShowCursor(TRUE);
			if ( zbufsurf ) { zbufsurf->Release(); zbufsurf=0; }
			if ( backsurf ) { backsurf->Release(); backsurf=0; }
			if ( primsurf ) { primsurf->Release(); primsurf=0; }
			if ( ddraw ) { ddraw->Release(); ddraw=0; }
			if ( rm ) { rm->Release(); rm=0; }
			return FALSE;
		}
	}

	device->SetQuality(D3DRMRENDER_GOURAUD);

	switch ( yes_full==1?bit:bpp )
	{
	case 16:
		device->SetShades( 32 );
		rm->SetDefaultTextureColors( 64 );
		rm->SetDefaultTextureShades( 32 );
		device->SetDither( FALSE );
		break;
	case 24:
	case 32:
		device->SetShades( 256 );
		rm->SetDefaultTextureColors( 64 );
		rm->SetDefaultTextureShades( 256 );
		device->SetDither( FALSE );
		break;
	}

	rc=rm->CreateFrame(NULL,&scene);
	if ( rc!=D3DRM_OK ) return FALSE;

	rc=CreateScene();
	if ( rc!=TRUE ) return FALSE;

	yes_render=1;

	return TRUE;
}


void	DestroyD3DRM()
{

	if ( yes_render==0 ) return;

	yes_render=0;

	DestroyScene();

	if ( scene ) { scene->Release(); scene=0; }

	if ( yes_full==1 )
	{
		ShowCursor(TRUE);
		if ( device ) { device->Release(); device=0; }
		if ( zbufsurf ) { zbufsurf->Release(); zbufsurf=0; }
		if ( backsurf ) { backsurf->Release(); backsurf=0; }
		if ( primsurf ) { primsurf->Release(); primsurf=0; }
		if ( ddraw ) { ddraw->Release(); ddraw=0; }
		if ( rm ) { rm->Release(); rm=0; }
	}
	else
	{
		if ( device ) { device->Release(); device=0; }
		if ( clip ) { clip->Release(); clip=0; }
		if ( rm ) { rm->Release(); rm=0; }
	}
}



void	ActivateD3DRM(WPARAM wParam)
{
	if ( yes_render==0 ) return;
	if ( yes_full==1 ) return;

	LPDIRECT3DRMWINDEVICE   win_dev;
	if ( device==0 ) return;
	if ( device->QueryInterface(IID_IDirect3DRMWinDevice,(void **)&win_dev)!=D3DRM_OK ) return;
	win_dev->HandleActivate((WORD)wParam);
	win_dev->Release();
}


void	RenderD3DRM()
{
	if ( yes_render==0 ) return;
	if ( yes_full==0 )
	{
		rm->Tick(D3DVALUE(1));
	}
	else
	{
		if ( primsurf->IsLost() == DDERR_SURFACELOST ) primsurf->Restore();
		scene->Move(D3DVALUE(1.0));
		viewport->Clear();
		viewport->ForceUpdate(0,0,rect_win.right-1,rect_win.bottom-1);
		viewport->Render( scene );
		device->Update();
		primsurf->Flip(0,DDFLIP_WAIT);
	}
}


void	PaintD3DRM()
{
	if ( yes_render==0 ) return;
	if ( yes_full==1 ) return;

	HDC hdc=GetDC(hwnd);
	int bpp_ptr=GetDeviceCaps(hdc,BITSPIXEL);
	ReleaseDC(hwnd,hdc);

	if ( bpp!=bpp_ptr )
	{
		DestroyD3DRM();
		InitD3DRM(hwnd,yes_full,0,0,0,yes_hardware);
	}

	RECT					rect;
	PAINTSTRUCT             ps;
	LPDIRECT3DRMWINDEVICE   win_dev;

	if ( GetUpdateRect(hwnd,&rect,FALSE)==FALSE ) return;
	BeginPaint( hwnd,&ps );
	if ( device->QueryInterface(IID_IDirect3DRMWinDevice,(void **)&win_dev)!=D3DRM_OK ) return;
	win_dev->HandlePaint(ps.hdc);
	win_dev->Release();
	EndPaint( hwnd,&ps );
}


void	SizeD3DRM(WPARAM wParam,LPARAM lParam)
{
	if ( yes_render==0 ) return;
	if ( yes_full==1 ) return;

	int w = LOWORD(lParam);
	int h = HIWORD(lParam);

	if ( (w==rect_win.right) && (h==rect_win.bottom) ) return;

	DestroyD3DRM();
	InitD3DRM(hwnd,yes_full,0,0,0,yes_hardware);
}


BOOL CreateScene()
{
	LPCSTR searchpath = "C:\\tmp\\samples\\multimedia\\media";
	rm->AddSearchPath(searchpath);

	HRESULT rc;
	rm->CreateMeshBuilder(&meshbuilder);
	rc = meshbuilder->Load("test.x",NULL,D3DRMLOAD_FROMFILE,NULL,NULL);
	if ( rc!=D3DRM_OK ) return FALSE;
	ScaleMesh(meshbuilder,D3DVALUE(25));

	LPDIRECT3DRMFRAME meshframe;
	rm->CreateFrame(scene,&meshframe);
	meshframe->SetRotation( scene,
			D3DVALUE(0), D3DVALUE(1), D3DVALUE(0),
			D3DVALUE(.1) );
	meshframe->AddVisual(meshbuilder);
	meshframe->Release();meshframe=0;

	LPDIRECT3DRMLIGHT slight;
	rm->CreateLightRGB(D3DRMLIGHT_SPOT,
		D3DVALUE(1.0),D3DVALUE(0.7),D3DVALUE(0.5),&slight);

	LPDIRECT3DRMFRAME slightframe;
	rm->CreateFrame(scene,&slightframe);
	slightframe->SetPosition(scene,D3DVALUE(0.),D3DVALUE(0.),D3DVALUE(-100.));
	slightframe->AddLight(slight);
	slightframe->Release();slightframe=0;
	slight->Release();slight=0;

	rm->CreateFrame(scene,&camera);
	camera->SetPosition(scene,D3DVALUE(0),D3DVALUE(30),D3DVALUE(-50));
	camera->SetOrientation(scene,
		D3DVALUE(0.),D3DVALUE(-0.7),D3DVALUE(1.),
		D3DVALUE(0.),D3DVALUE(1.),D3DVALUE(0.));

	rm->CreateViewport(device,camera,0,0,
		device->GetWidth(),device->GetHeight(),&viewport);

	return TRUE;
}


void ScaleMesh(LPDIRECT3DRMMESHBUILDER mesh, D3DVALUE dim)
{
	D3DRMBOX box;
	mesh->GetBox(&box);	
	D3DVALUE sizex = box.max.x - box.min.x;
	D3DVALUE sizey = box.max.y - box.min.y;
	D3DVALUE sizez = box.max.z - box.min.z;
	D3DVALUE largedim = D3DVALUE(0);	
	if ( sizex>largedim ) largedim=sizex;
	if ( sizey>largedim ) largedim=sizey;
	if ( sizez>largedim ) largedim=sizez;	
	D3DVALUE scalefactor = dim/largedim;
	mesh->Scale(scalefactor,scalefactor,scalefactor);
}


void DestroyScene()
{
	if ( camera ) camera->Release();camera=0;
	if ( viewport ) viewport->Release();viewport=0;
	if ( meshbuilder ) meshbuilder->Release();meshbuilder=0;
}

